
#ifndef _arrays_
#define _arrays_

#include "redblack.h"

/*-
   Lists are implemented as arrays which may grow from the right or left.
   The interesting indexes are shown in the diagram below:

            base (pointer)
            |
   |________|___________________|_________________|
   left  active_left        active_right        right

   There is just one pointer, the base pointer.  The other
   indexes are integers.

   Basically, the routines manipulate active_left and
   active_right, and ensure that they remain bounded by
   left and right.

   As shown, the indexes are over-specified; base may be chosen to
   coincide with either left or active_left.  These routines choose to
   make base and active_left coincide; ie, active_left is defined to be
   zero.  This makes the grow routine more complex, but the set and get
   routines less complex.

   Each entry of the list has a particular size, which is often
   sizeof (char *).  Correspondingly, while the list is often
   an array of pointers, it may be an array of structures or
   doubles, integers, shorts, or chars.
*/

#define TBL_MIN_INCREMENT 10

typedef struct Tbl
{
    union     { 
	char *cp ; 
	char **pp ; 
	} base ;
    char	   *mapped ; 
    char           *null;
    char	   *deleted;
    long            left,
		    active_right,
                    right;
    int             entry_size;
    int		    direct ;
    long	    mapped_size ;
    int		    fd; 
    int		    prot ;
    long	    left_grow, right_grow ;
}               Tbl;

typedef Tbl     Bitvector;

typedef struct Stbl
{
    Tbl            *tbl;
    Arr		   *arr;
    int             changed;
}               Stbl;

#define maxtbl(TBL) ((TBL)->active_right)
#define sizetbl(TBL) ((TBL)->right) 
#define NULLTBL { {0}, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
#define EMPTYTBL(DIRECT,ENTRYSIZE,RIGHTGROW) { {0}, 0, 0, 0, 0, 0, 0, ENTRYSIZE, DIRECT, 0, 0, 0, 0, RIGHTGROW }

#ifdef  __cplusplus
extern "C" {
#endif

extern Arr *newarr (int (*cmp)(char *, char *)) ;
extern void *delarr ( Arr *arr, char *key );
extern void *setarr ( Arr *arr, char *key, void *value );
extern void *getarr ( Arr *arr, char *key );
extern Tbl *keysarr ( Arr *arr );
extern Tbl *valsarr ( Arr *arr );
extern int freearr (Arr * root, void (*afree) (void *)) ;
extern long cntarr ( Arr *arr );
extern int applyarr ( Arr * arr, void (*function)(char *key, void *value, void *pvt), void *);
extern Arr * duparr ( Arr *arr, void *(*adup)(void *) );

void diffarr ( Arr *a, Arr *b, int (*cmpkey)(const char *, const char *, const void *), int (*cmpval)(const void *, const void *, const void *), const void *pvt, Tbl *a_only, Tbl *b_only, Tbl *different ) ;
extern Tbl * greparr ( char *re, Arr *arr );
extern Tbl * arrgrep ( Arr *arr, char *string );

extern Tbl * inittbl ( long left_grow, long right_grow, int direct, void *null, int entry_size );
extern Tbl * newtbl ( long size );
extern long settbl ( Tbl * tbl, long index, void *value );
extern void * addrtbl ( Tbl * tbl, long index );
extern void * gettbl ( Tbl * tbl, long index );
extern void freetbl ( Tbl * tbl, void (*free_tblval)(void *) );
extern void trunctbl ( Tbl * tbl, long last, void (*free_tblval) (void *) );
extern void clrtbl ( Tbl * tbl, void (*free_tblval)(void *) );
extern int applytbl ( Tbl * tbl, int (*function) (void *, void *), void *pvt );
extern long pushtbl ( Tbl * tbl, void *value );
extern void * poptbl ( Tbl * tbl );
extern long unshifttbl ( Tbl * tbl, void *value );
extern void * shifttbl ( Tbl * tbl );
extern long instbl ( Tbl * tbl, long index, void *value );
extern void * deltbl ( Tbl * tbl, long index );
extern long cattbl ( Tbl * tbldest, Tbl * tblsrc );
extern int writetbl ( Tbl * tbl, FILE * file, char *s );
extern int savetbl ( Tbl * tbl, char *filename, char *s );
extern Tbl * readtbl ( char *filename, char **s, int rw );
extern int filetbl ( char *filename, char *identifier, long n, void *null, int entry_size, Tbl ** ptbl );
extern int cmptbl ( Tbl * tbl1, Tbl * tbl2 );
extern long searchtbl ( char *key, Tbl * tbl, int (*compare)(char *, char *, void *), void *pvt, long *ns, long *ne );
extern void sorttbl ( Tbl * tbl, int (*compare) (char **, char **, void *), void *pvt );
extern Tbl * duptbl ( Tbl * old, void *(*dup_tblval) (void *) );
extern Tbl * strtbl ( char *s, ... );
extern char * jointbl ( Tbl * tbl, char *separator );
extern Tbl * greptbl ( char *re, Tbl * tbl );
extern Tbl * tblgrep ( Tbl * tbl, char *string );
extern void dumptbl (char *msg, Tbl *list, int (*prt)(const char *)) ;

void difftbl ( Tbl *a, Tbl *b, int (*cmp)(const char *, const char *, const void *), const void *pvt, Tbl *a_only, Tbl *b_only, Tbl *both ) ; 

extern Stbl * newstbl ( int (*cmp) ( void *, void *) );
extern void * addstbl ( Stbl * stbl, void *key );
extern void * delstbl ( Stbl * stbl, void *key );
extern void * tststbl ( Stbl * stbl, void *key );
extern long maxstbl ( Stbl * stbl );
extern void * getstbl ( Stbl * stbl, long index );
extern Tbl * tblstbl ( Stbl * stbl );
extern int freestbl ( Stbl * stbl, void (*free_stblval)(void *) );
extern int applystbl ( Stbl * stbl, int (*function)(void *, void *), void *pvt );

extern Stbl * newidmap ( void );
extern void freeidmap ( Stbl *stbl);
extern long idset ( Stbl *stbl, long ix, long iy);
extern long idget ( Stbl *stbl, long ix);

#ifdef	__cplusplus
}
#endif

#endif

/* $Id$ */
